"""
Tracking visualizer for SBYB.

This module provides the TrackingVisualizer class, which generates visualizations
for experiment tracking data to help users understand and compare their experiments.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import json
import datetime
import uuid
from pathlib import Path
import logging
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import TrackingError
from sbyb.tracking.experiment import Experiment
from sbyb.tracking.run import Run
from sbyb.tracking.storage import LocalStorage


class TrackingVisualizer(SBYBComponent):
    """
    Visualizer for experiment tracking data.
    
    This component provides functionality for generating visualizations
    of experiment tracking data to help users understand and compare
    their experiments.
    """
    
    def __init__(self, storage: Optional[LocalStorage] = None, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the tracking visualizer.
        
        Args:
            storage: Optional LocalStorage instance. If None, a new instance is created.
            config: Optional configuration dictionary.
        """
        super().__init__(config)
        
        # Set up logging
        self.logger = logging.getLogger("sbyb.tracking.visualizer")
        
        # Initialize storage
        self.storage = storage or LocalStorage(config=config)
        
        # Set default style for matplotlib
        plt.style.use('seaborn-v0_8-whitegrid')
    
    def plot_metric_history(self, run_id: str, metric: str, 
                           figsize: Tuple[int, int] = (10, 6),
                           title: Optional[str] = None,
                           use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Plot the history of a metric for a run.
        
        Args:
            run_id: ID of the run.
            metric: Name of the metric to plot.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
            
        Raises:
            TrackingError: If the run is not found or the metric has no history.
        """
        run = self.storage.load_run(run_id)
        
        if metric not in run.history:
            raise TrackingError(f"Metric {metric} has no history for run {run_id}")
        
        history = run.history[metric]
        steps = [entry["step"] for entry in history]
        values = [entry["value"] for entry in history]
        
        if use_plotly:
            fig = px.line(
                x=steps,
                y=values,
                labels={"x": "Step", "y": metric},
                title=title or f"{metric} History for {run.name}"
            )
            fig.update_layout(
                xaxis_title="Step",
                yaxis_title=metric,
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            ax.plot(steps, values, marker='o', linestyle='-', markersize=4)
            ax.set_xlabel("Step")
            ax.set_ylabel(metric)
            ax.set_title(title or f"{metric} History for {run.name}")
            ax.grid(True)
            return fig
    
    def plot_metric_comparison(self, run_ids: List[str], metric: str,
                              figsize: Tuple[int, int] = (12, 6),
                              title: Optional[str] = None,
                              use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Compare a metric across multiple runs.
        
        Args:
            run_ids: List of run IDs to compare.
            metric: Name of the metric to compare.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
            
        Raises:
            TrackingError: If any of the runs are not found or the metric is not found.
        """
        runs = []
        for run_id in run_ids:
            try:
                run = self.storage.load_run(run_id)
                runs.append(run)
            except TrackingError:
                self.logger.warning(f"Run not found: {run_id}")
        
        if not runs:
            raise TrackingError("No valid runs found")
        
        # Check if the metric exists in any run
        if not any(metric in run.metrics for run in runs):
            raise TrackingError(f"Metric {metric} not found in any run")
        
        # Create data for plotting
        run_names = []
        metric_values = []
        
        for run in runs:
            run_names.append(run.name)
            metric_values.append(run.metrics.get(metric, np.nan))
        
        if use_plotly:
            fig = px.bar(
                x=run_names,
                y=metric_values,
                labels={"x": "Run", "y": metric},
                title=title or f"{metric} Comparison"
            )
            fig.update_layout(
                xaxis_title="Run",
                yaxis_title=metric,
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            bars = ax.bar(run_names, metric_values)
            
            # Add value labels on top of bars
            for bar, value in zip(bars, metric_values):
                if not np.isnan(value):
                    height = bar.get_height()
                    ax.text(bar.get_x() + bar.get_width()/2., height + 0.01,
                           f'{value:.4f}', ha='center', va='bottom', rotation=0)
            
            ax.set_xlabel("Run")
            ax.set_ylabel(metric)
            ax.set_title(title or f"{metric} Comparison")
            plt.xticks(rotation=45, ha='right')
            plt.tight_layout()
            return fig
    
    def plot_metric_correlation(self, experiment_id: str, metrics: List[str],
                               figsize: Tuple[int, int] = (10, 8),
                               title: Optional[str] = None,
                               use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Plot correlation between metrics across runs in an experiment.
        
        Args:
            experiment_id: ID of the experiment.
            metrics: List of metrics to correlate.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
            
        Raises:
            TrackingError: If the experiment is not found or has no runs.
        """
        experiment = self.storage.load_experiment(experiment_id)
        
        if not experiment.runs:
            raise TrackingError(f"Experiment {experiment_id} has no runs")
        
        # Create a DataFrame with metrics from all runs
        data = []
        for run in experiment.runs:
            run_data = {"run_id": run.run_id, "run_name": run.name}
            for metric in metrics:
                run_data[metric] = run.metrics.get(metric, np.nan)
            data.append(run_data)
        
        df = pd.DataFrame(data)
        
        # Check if we have enough data
        if len(df) < 2:
            raise TrackingError("Need at least 2 runs for correlation analysis")
        
        # Calculate correlation matrix
        corr_matrix = df[metrics].corr()
        
        if use_plotly:
            fig = px.imshow(
                corr_matrix,
                text_auto=True,
                color_continuous_scale='RdBu_r',
                title=title or f"Metric Correlation for {experiment.name}"
            )
            fig.update_layout(
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            mask = np.triu(np.ones_like(corr_matrix, dtype=bool))
            cmap = sns.diverging_palette(230, 20, as_cmap=True)
            
            sns.heatmap(corr_matrix, mask=mask, cmap=cmap, vmax=1, vmin=-1, center=0,
                       square=True, linewidths=.5, annot=True, fmt=".2f", ax=ax)
            
            ax.set_title(title or f"Metric Correlation for {experiment.name}")
            plt.tight_layout()
            return fig
    
    def plot_parameter_importance(self, experiment_id: str, metric: str,
                                 figsize: Tuple[int, int] = (12, 8),
                                 title: Optional[str] = None,
                                 use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Plot parameter importance for a metric across runs in an experiment.
        
        Args:
            experiment_id: ID of the experiment.
            metric: Metric to analyze.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
            
        Raises:
            TrackingError: If the experiment is not found or has no runs.
        """
        experiment = self.storage.load_experiment(experiment_id)
        
        if not experiment.runs:
            raise TrackingError(f"Experiment {experiment_id} has no runs")
        
        # Create a DataFrame with parameters and the metric from all runs
        data = []
        for run in experiment.runs:
            if metric in run.metrics:
                run_data = {"run_id": run.run_id, "run_name": run.name, metric: run.metrics[metric]}
                for param_name, param_value in run.parameters.items():
                    # Only include numeric or boolean parameters
                    if isinstance(param_value, (int, float, bool)):
                        run_data[param_name] = param_value
                data.append(run_data)
        
        df = pd.DataFrame(data)
        
        # Check if we have enough data
        if len(df) < 3:
            raise TrackingError("Need at least 3 runs for parameter importance analysis")
        
        # Get parameter columns
        param_cols = [col for col in df.columns if col not in ["run_id", "run_name", metric]]
        
        if not param_cols:
            raise TrackingError("No numeric parameters found in runs")
        
        # Calculate correlation between parameters and the metric
        correlations = []
        for param in param_cols:
            if df[param].nunique() > 1:  # Only include parameters with variation
                corr = df[[param, metric]].corr().iloc[0, 1]
                correlations.append((param, abs(corr), corr))
        
        if not correlations:
            raise TrackingError("No parameters with variation found")
        
        # Sort by absolute correlation
        correlations.sort(key=lambda x: x[1], reverse=True)
        
        # Extract data for plotting
        params = [c[0] for c in correlations]
        abs_corrs = [c[1] for c in correlations]
        corrs = [c[2] for c in correlations]
        
        if use_plotly:
            fig = px.bar(
                x=params,
                y=abs_corrs,
                color=corrs,
                color_continuous_scale='RdBu_r',
                labels={"x": "Parameter", "y": f"Correlation with {metric} (absolute)"},
                title=title or f"Parameter Importance for {metric}"
            )
            fig.update_layout(
                xaxis_title="Parameter",
                yaxis_title=f"Correlation with {metric} (absolute)",
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            bars = ax.bar(params, abs_corrs, color=plt.cm.RdBu_r(np.array(corrs) * 0.5 + 0.5))
            
            # Add value labels on top of bars
            for bar, value in zip(bars, corrs):
                height = bar.get_height()
                ax.text(bar.get_x() + bar.get_width()/2., height + 0.01,
                       f'{value:.2f}', ha='center', va='bottom', rotation=0)
            
            ax.set_xlabel("Parameter")
            ax.set_ylabel(f"Correlation with {metric} (absolute)")
            ax.set_title(title or f"Parameter Importance for {metric}")
            plt.xticks(rotation=45, ha='right')
            plt.tight_layout()
            return fig
    
    def plot_parallel_coordinates(self, experiment_id: str, params: List[str], metrics: List[str],
                                 figsize: Tuple[int, int] = (14, 8),
                                 title: Optional[str] = None) -> go.Figure:
        """
        Create a parallel coordinates plot for parameters and metrics.
        
        Args:
            experiment_id: ID of the experiment.
            params: List of parameters to include.
            metrics: List of metrics to include.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            
        Returns:
            Plotly Figure object.
            
        Raises:
            TrackingError: If the experiment is not found or has no runs.
        """
        experiment = self.storage.load_experiment(experiment_id)
        
        if not experiment.runs:
            raise TrackingError(f"Experiment {experiment_id} has no runs")
        
        # Create a DataFrame with parameters and metrics from all runs
        data = []
        for run in experiment.runs:
            run_data = {"run_id": run.run_id, "run_name": run.name}
            
            # Add parameters
            for param in params:
                run_data[f"param_{param}"] = run.parameters.get(param, np.nan)
            
            # Add metrics
            for metric in metrics:
                run_data[f"metric_{metric}"] = run.metrics.get(metric, np.nan)
            
            data.append(run_data)
        
        df = pd.DataFrame(data)
        
        # Check if we have enough data
        if len(df) < 2:
            raise TrackingError("Need at least 2 runs for parallel coordinates plot")
        
        # Create dimensions for the plot
        dimensions = []
        
        # Add parameter dimensions
        for param in params:
            col = f"param_{param}"
            if col in df.columns and df[col].nunique() > 1:
                dimensions.append(
                    dict(
                        range=[df[col].min(), df[col].max()],
                        label=param,
                        values=df[col]
                    )
                )
        
        # Add metric dimensions
        for metric in metrics:
            col = f"metric_{metric}"
            if col in df.columns and df[col].nunique() > 1:
                dimensions.append(
                    dict(
                        range=[df[col].min(), df[col].max()],
                        label=metric,
                        values=df[col]
                    )
                )
        
        if not dimensions:
            raise TrackingError("No valid dimensions found for parallel coordinates plot")
        
        # Create the plot
        fig = go.Figure(data=
            go.Parcoords(
                line=dict(color=df.index, colorscale='Viridis', showscale=True),
                dimensions=dimensions
            )
        )
        
        fig.update_layout(
            title=title or f"Parallel Coordinates for {experiment.name}",
            width=figsize[0] * 100,
            height=figsize[1] * 100,
            template="plotly_white"
        )
        
        return fig
    
    def plot_run_status_distribution(self, experiment_id: Optional[str] = None,
                                    figsize: Tuple[int, int] = (10, 6),
                                    title: Optional[str] = None,
                                    use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Plot the distribution of run statuses.
        
        Args:
            experiment_id: Optional ID of the experiment to filter by.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
        """
        # Get run data
        runs = self.storage.list_runs(experiment_id)
        
        if not runs:
            if experiment_id:
                raise TrackingError(f"Experiment {experiment_id} has no runs")
            else:
                raise TrackingError("No runs found")
        
        # Count statuses
        status_counts = {}
        for run in runs:
            status = run.get("status", "unknown")
            status_counts[status] = status_counts.get(status, 0) + 1
        
        statuses = list(status_counts.keys())
        counts = list(status_counts.values())
        
        if use_plotly:
            fig = px.pie(
                names=statuses,
                values=counts,
                title=title or "Run Status Distribution"
            )
            fig.update_layout(
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            ax.pie(counts, labels=statuses, autopct='%1.1f%%', startangle=90)
            ax.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle
            ax.set_title(title or "Run Status Distribution")
            return fig
    
    def plot_metric_distribution(self, experiment_id: str, metric: str,
                               figsize: Tuple[int, int] = (10, 6),
                               title: Optional[str] = None,
                               use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Plot the distribution of a metric across runs in an experiment.
        
        Args:
            experiment_id: ID of the experiment.
            metric: Metric to analyze.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
            
        Raises:
            TrackingError: If the experiment is not found or has no runs.
        """
        experiment = self.storage.load_experiment(experiment_id)
        
        if not experiment.runs:
            raise TrackingError(f"Experiment {experiment_id} has no runs")
        
        # Extract metric values
        values = []
        for run in experiment.runs:
            if metric in run.metrics:
                values.append(run.metrics[metric])
        
        if not values:
            raise TrackingError(f"Metric {metric} not found in any run")
        
        if use_plotly:
            fig = px.histogram(
                x=values,
                nbins=min(20, len(values)),
                labels={"x": metric},
                title=title or f"{metric} Distribution"
            )
            fig.update_layout(
                xaxis_title=metric,
                yaxis_title="Count",
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            ax.hist(values, bins=min(20, len(values)), alpha=0.7)
            ax.set_xlabel(metric)
            ax.set_ylabel("Count")
            ax.set_title(title or f"{metric} Distribution")
            return fig
    
    def plot_run_duration(self, experiment_id: Optional[str] = None,
                         figsize: Tuple[int, int] = (12, 6),
                         title: Optional[str] = None,
                         use_plotly: bool = False) -> Union[Figure, go.Figure]:
        """
        Plot the duration of runs.
        
        Args:
            experiment_id: Optional ID of the experiment to filter by.
            figsize: Figure size (width, height) in inches.
            title: Optional title for the plot.
            use_plotly: Whether to use Plotly instead of Matplotlib.
            
        Returns:
            Matplotlib Figure or Plotly Figure object.
        """
        # Get run data
        runs = self.storage.list_runs(experiment_id)
        
        if not runs:
            if experiment_id:
                raise TrackingError(f"Experiment {experiment_id} has no runs")
            else:
                raise TrackingError("No runs found")
        
        # Calculate durations
        run_names = []
        durations = []
        
        for run in runs:
            if run.get("started_at") and run.get("ended_at"):
                run_names.append(run.get("name", run.get("run_id")))
                
                started = datetime.datetime.fromisoformat(run["started_at"])
                ended = datetime.datetime.fromisoformat(run["ended_at"])
                duration = (ended - started).total_seconds() / 60.0  # Duration in minutes
                
                durations.append(duration)
        
        if not durations:
            raise TrackingError("No runs with valid duration information found")
        
        # Sort by duration
        sorted_indices = np.argsort(durations)
        sorted_names = [run_names[i] for i in sorted_indices]
        sorted_durations = [durations[i] for i in sorted_indices]
        
        if use_plotly:
            fig = px.bar(
                x=sorted_names,
                y=sorted_durations,
                labels={"x": "Run", "y": "Duration (minutes)"},
                title=title or "Run Duration"
            )
            fig.update_layout(
                xaxis_title="Run",
                yaxis_title="Duration (minutes)",
                template="plotly_white"
            )
            return fig
        else:
            fig, ax = plt.subplots(figsize=figsize)
            ax.bar(sorted_names, sorted_durations)
            ax.set_xlabel("Run")
            ax.set_ylabel("Duration (minutes)")
            ax.set_title(title or "Run Duration")
            plt.xticks(rotation=45, ha='right')
            plt.tight_layout()
            return fig
    
    def create_experiment_dashboard(self, experiment_id: str, 
                                   metrics: List[str],
                                   params: Optional[List[str]] = None) -> go.Figure:
        """
        Create a comprehensive dashboard for an experiment.
        
        Args:
            experiment_id: ID of the experiment.
            metrics: List of metrics to include.
            params: Optional list of parameters to include.
            
        Returns:
            Plotly Figure object with multiple subplots.
            
        Raises:
            TrackingError: If the experiment is not found or has no runs.
        """
        experiment = self.storage.load_experiment(experiment_id)
        
        if not experiment.runs:
            raise TrackingError(f"Experiment {experiment_id} has no runs")
        
        # Create a subplot grid
        fig = make_subplots(
            rows=2, cols=2,
            subplot_titles=(
                "Metric Comparison",
                "Run Status Distribution",
                "Parameter Importance",
                "Run Duration"
            ),
            specs=[
                [{"type": "bar"}, {"type": "pie"}],
                [{"type": "bar"}, {"type": "bar"}]
            ]
        )
        
        # 1. Metric Comparison
        run_names = []
        metric_values = {metric: [] for metric in metrics}
        
        for run in experiment.runs:
            run_names.append(run.name)
            for metric in metrics:
                metric_values[metric].append(run.metrics.get(metric, np.nan))
        
        # Add traces for each metric
        for i, metric in enumerate(metrics):
            fig.add_trace(
                go.Bar(
                    x=run_names,
                    y=metric_values[metric],
                    name=metric
                ),
                row=1, col=1
            )
        
        # 2. Run Status Distribution
        status_counts = {}
        for run in experiment.runs:
            status = run.status
            status_counts[status] = status_counts.get(status, 0) + 1
        
        statuses = list(status_counts.keys())
        counts = list(status_counts.values())
        
        fig.add_trace(
            go.Pie(
                labels=statuses,
                values=counts,
                textinfo='percent+label'
            ),
            row=1, col=2
        )
        
        # 3. Parameter Importance (if parameters are provided)
        if params:
            # Get the first metric for parameter importance
            metric = metrics[0]
            
            # Create a DataFrame with parameters and the metric
            data = []
            for run in experiment.runs:
                if metric in run.metrics:
                    run_data = {"run_id": run.run_id, "run_name": run.name, metric: run.metrics[metric]}
                    for param_name in params:
                        param_value = run.parameters.get(param_name)
                        if isinstance(param_value, (int, float, bool)):
                            run_data[param_name] = param_value
                    data.append(run_data)
            
            df = pd.DataFrame(data)
            
            # Calculate correlation between parameters and the metric
            correlations = []
            for param in params:
                if param in df.columns and df[param].nunique() > 1:
                    corr = df[[param, metric]].corr().iloc[0, 1]
                    if not np.isnan(corr):
                        correlations.append((param, abs(corr), corr))
            
            if correlations:
                # Sort by absolute correlation
                correlations.sort(key=lambda x: x[1], reverse=True)
                
                # Extract data for plotting
                params_sorted = [c[0] for c in correlations]
                abs_corrs = [c[1] for c in correlations]
                
                fig.add_trace(
                    go.Bar(
                        x=params_sorted,
                        y=abs_corrs,
                        name="Parameter Importance"
                    ),
                    row=2, col=1
                )
        
        # 4. Run Duration
        run_names = []
        durations = []
        
        for run in experiment.runs:
            if run.started_at and run.ended_at:
                run_names.append(run.name)
                
                started = datetime.datetime.fromisoformat(run.started_at)
                ended = datetime.datetime.fromisoformat(run.ended_at)
                duration = (ended - started).total_seconds() / 60.0  # Duration in minutes
                
                durations.append(duration)
        
        if durations:
            # Sort by duration
            sorted_indices = np.argsort(durations)
            sorted_names = [run_names[i] for i in sorted_indices]
            sorted_durations = [durations[i] for i in sorted_indices]
            
            fig.add_trace(
                go.Bar(
                    x=sorted_names,
                    y=sorted_durations,
                    name="Duration (minutes)"
                ),
                row=2, col=2
            )
        
        # Update layout
        fig.update_layout(
            title_text=f"Dashboard for {experiment.name}",
            height=800,
            width=1200,
            template="plotly_white",
            showlegend=False
        )
        
        return fig
    
    def export_experiment_report(self, experiment_id: str, output_dir: str,
                               metrics: List[str], params: Optional[List[str]] = None) -> str:
        """
        Export a comprehensive HTML report for an experiment.
        
        Args:
            experiment_id: ID of the experiment.
            output_dir: Directory to save the report to.
            metrics: List of metrics to include.
            params: Optional list of parameters to include.
            
        Returns:
            Path to the generated HTML report.
            
        Raises:
            TrackingError: If the experiment is not found or has no runs.
        """
        experiment = self.storage.load_experiment(experiment_id)
        
        if not experiment.runs:
            raise TrackingError(f"Experiment {experiment_id} has no runs")
        
        # Create output directory
        os.makedirs(output_dir, exist_ok=True)
        
        # Generate plots
        plots = {}
        
        try:
            # Dashboard
            plots["dashboard"] = self.create_experiment_dashboard(
                experiment_id=experiment_id,
                metrics=metrics,
                params=params
            )
            
            # Metric comparison for each metric
            for metric in metrics:
                try:
                    plots[f"metric_{metric}"] = self.plot_metric_comparison(
                        run_ids=[run.run_id for run in experiment.runs],
                        metric=metric,
                        use_plotly=True
                    )
                except Exception as e:
                    self.logger.warning(f"Error generating metric comparison for {metric}: {str(e)}")
            
            # Parameter importance for each metric
            if params:
                for metric in metrics:
                    try:
                        plots[f"param_importance_{metric}"] = self.plot_parameter_importance(
                            experiment_id=experiment_id,
                            metric=metric,
                            use_plotly=True
                        )
                    except Exception as e:
                        self.logger.warning(f"Error generating parameter importance for {metric}: {str(e)}")
            
            # Parallel coordinates
            if params:
                try:
                    plots["parallel_coords"] = self.plot_parallel_coordinates(
                        experiment_id=experiment_id,
                        params=params,
                        metrics=metrics
                    )
                except Exception as e:
                    self.logger.warning(f"Error generating parallel coordinates plot: {str(e)}")
            
            # Run duration
            try:
                plots["run_duration"] = self.plot_run_duration(
                    experiment_id=experiment_id,
                    use_plotly=True
                )
            except Exception as e:
                self.logger.warning(f"Error generating run duration plot: {str(e)}")
            
            # Run status distribution
            try:
                plots["run_status"] = self.plot_run_status_distribution(
                    experiment_id=experiment_id,
                    use_plotly=True
                )
            except Exception as e:
                self.logger.warning(f"Error generating run status distribution plot: {str(e)}")
            
            # Generate HTML
            html_content = f"""
            <!DOCTYPE html>
            <html>
            <head>
                <title>Experiment Report: {experiment.name}</title>
                <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
                <style>
                    body {{
                        font-family: Arial, sans-serif;
                        margin: 0;
                        padding: 20px;
                        background-color: #f5f5f5;
                    }}
                    .container {{
                        max-width: 1200px;
                        margin: 0 auto;
                        background-color: white;
                        padding: 20px;
                        box-shadow: 0 0 10px rgba(0,0,0,0.1);
                    }}
                    h1, h2, h3 {{
                        color: #333;
                    }}
                    .plot {{
                        margin-bottom: 30px;
                    }}
                    table {{
                        width: 100%;
                        border-collapse: collapse;
                        margin-bottom: 20px;
                    }}
                    th, td {{
                        padding: 8px;
                        text-align: left;
                        border-bottom: 1px solid #ddd;
                    }}
                    th {{
                        background-color: #f2f2f2;
                    }}
                    tr:hover {{
                        background-color: #f5f5f5;
                    }}
                </style>
            </head>
            <body>
                <div class="container">
                    <h1>Experiment Report: {experiment.name}</h1>
                    <p><strong>Description:</strong> {experiment.description}</p>
                    <p><strong>Created:</strong> {experiment.created_at}</p>
                    <p><strong>Tags:</strong> {', '.join(experiment.tags) if experiment.tags else 'None'}</p>
                    <p><strong>Number of Runs:</strong> {len(experiment.runs)}</p>
                    
                    <h2>Dashboard</h2>
                    <div class="plot" id="dashboard"></div>
                    
                    <h2>Metrics</h2>
            """
            
            # Add metric plots
            for metric in metrics:
                plot_id = f"metric_{metric}"
                if plot_id in plots:
                    html_content += f"""
                    <h3>{metric}</h3>
                    <div class="plot" id="{plot_id}"></div>
                    """
            
            # Add parameter importance plots
            if params:
                html_content += "<h2>Parameter Importance</h2>"
                for metric in metrics:
                    plot_id = f"param_importance_{metric}"
                    if plot_id in plots:
                        html_content += f"""
                        <h3>For {metric}</h3>
                        <div class="plot" id="{plot_id}"></div>
                        """
            
            # Add parallel coordinates plot
            if "parallel_coords" in plots:
                html_content += """
                <h2>Parameter-Metric Relationships</h2>
                <div class="plot" id="parallel_coords"></div>
                """
            
            # Add run duration and status plots
            html_content += """
            <h2>Run Information</h2>
            """
            
            if "run_duration" in plots:
                html_content += """
                <h3>Run Duration</h3>
                <div class="plot" id="run_duration"></div>
                """
            
            if "run_status" in plots:
                html_content += """
                <h3>Run Status Distribution</h3>
                <div class="plot" id="run_status"></div>
                """
            
            # Add run table
            html_content += """
            <h2>Run Details</h2>
            <table>
                <tr>
                    <th>Name</th>
                    <th>Status</th>
                    <th>Created</th>
                    <th>Duration</th>
            """
            
            for metric in metrics:
                html_content += f"<th>{metric}</th>"
            
            if params:
                for param in params:
                    html_content += f"<th>{param}</th>"
            
            html_content += "</tr>"
            
            for run in experiment.runs:
                duration = ""
                if run.started_at and run.ended_at:
                    started = datetime.datetime.fromisoformat(run.started_at)
                    ended = datetime.datetime.fromisoformat(run.ended_at)
                    duration_mins = (ended - started).total_seconds() / 60.0
                    duration = f"{duration_mins:.2f} min"
                
                html_content += f"""
                <tr>
                    <td>{run.name}</td>
                    <td>{run.status}</td>
                    <td>{run.created_at}</td>
                    <td>{duration}</td>
                """
                
                for metric in metrics:
                    value = run.metrics.get(metric, "")
                    html_content += f"<td>{value}</td>"
                
                if params:
                    for param in params:
                        value = run.parameters.get(param, "")
                        html_content += f"<td>{value}</td>"
                
                html_content += "</tr>"
            
            html_content += """
                </table>
                </div>
            """
            
            # Add plot scripts
            html_content += "<script>"
            for plot_id, plot in plots.items():
                plot_json = plot.to_json()
                html_content += f"""
                Plotly.newPlot('{plot_id}', {plot_json});
                """
            html_content += "</script>"
            
            html_content += """
            </body>
            </html>
            """
            
            # Write HTML to file
            report_path = os.path.join(output_dir, f"experiment_{experiment_id}_report.html")
            with open(report_path, "w") as f:
                f.write(html_content)
            
            return report_path
        
        except Exception as e:
            raise TrackingError(f"Error generating experiment report: {str(e)}")



"""
Base classes and interfaces for SBYB.

This module defines the abstract base classes and interfaces that form the foundation
of the SBYB library's architecture.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd


class SBYBComponent(ABC):
    """Base class for all SBYB components."""
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the component.
        
        Args:
            config: Configuration dictionary for the component.
        """
        self.config = config or {}
        self._validate_config()
        
    def _validate_config(self) -> None:
        """Validate the configuration."""
        pass
    
    @property
    def name(self) -> str:
        """Return the name of the component."""
        return self.__class__.__name__


class DataProcessor(SBYBComponent):
    """Base class for all data processors."""
    
    @abstractmethod
    def fit(self, data: Union[pd.DataFrame, np.ndarray], **kwargs) -> 'DataProcessor':
        """
        Fit the processor to the data.
        
        Args:
            data: Input data to fit the processor.
            **kwargs: Additional keyword arguments.
            
        Returns:
            self: The fitted processor.
        """
        pass
    
    @abstractmethod
    def transform(self, data: Union[pd.DataFrame, np.ndarray], **kwargs) -> Union[pd.DataFrame, np.ndarray]:
        """
        Transform the data using the fitted processor.
        
        Args:
            data: Input data to transform.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Transformed data.
        """
        pass
    
    def fit_transform(self, data: Union[pd.DataFrame, np.ndarray], **kwargs) -> Union[pd.DataFrame, np.ndarray]:
        """
        Fit the processor to the data and transform it.
        
        Args:
            data: Input data to fit and transform.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Transformed data.
        """
        return self.fit(data, **kwargs).transform(data, **kwargs)


class Model(SBYBComponent):
    """Base class for all models."""
    
    @abstractmethod
    def fit(self, X: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray], **kwargs) -> 'Model':
        """
        Fit the model to the data.
        
        Args:
            X: Input features.
            y: Target variable.
            **kwargs: Additional keyword arguments.
            
        Returns:
            self: The fitted model.
        """
        pass
    
    @abstractmethod
    def predict(self, X: Union[pd.DataFrame, np.ndarray], **kwargs) -> Union[pd.Series, np.ndarray]:
        """
        Make predictions using the fitted model.
        
        Args:
            X: Input features.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Predictions.
        """
        pass
    
    def score(self, X: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray], **kwargs) -> float:
        """
        Calculate the score of the model on the given data.
        
        Args:
            X: Input features.
            y: Target variable.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Score value.
        """
        pass


class Pipeline(SBYBComponent):
    """Base class for pipelines that combine multiple components."""
    
    def __init__(self, steps: List[Tuple[str, SBYBComponent]], config: Optional[Dict[str, Any]] = None):
        """
        Initialize the pipeline.
        
        Args:
            steps: List of (name, component) tuples defining the pipeline steps.
            config: Configuration dictionary for the pipeline.
        """
        super().__init__(config)
        self.steps = steps
        
    @abstractmethod
    def fit(self, data: Any, **kwargs) -> 'Pipeline':
        """
        Fit the pipeline to the data.
        
        Args:
            data: Input data.
            **kwargs: Additional keyword arguments.
            
        Returns:
            self: The fitted pipeline.
        """
        pass
    
    @abstractmethod
    def transform(self, data: Any, **kwargs) -> Any:
        """
        Transform the data using the fitted pipeline.
        
        Args:
            data: Input data.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Transformed data.
        """
        pass
    
    def fit_transform(self, data: Any, **kwargs) -> Any:
        """
        Fit the pipeline to the data and transform it.
        
        Args:
            data: Input data.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Transformed data.
        """
        return self.fit(data, **kwargs).transform(data, **kwargs)


class TaskDetector(SBYBComponent):
    """Base class for task detection components."""
    
    @abstractmethod
    def detect(self, data: Union[pd.DataFrame, np.ndarray], target: Optional[Union[str, int]] = None) -> str:
        """
        Detect the machine learning task type from the data.
        
        Args:
            data: Input data.
            target: Target variable name or index.
            
        Returns:
            Detected task type.
        """
        pass


class Evaluator(SBYBComponent):
    """Base class for model evaluation components."""
    
    @abstractmethod
    def evaluate(self, model: Model, X: Union[pd.DataFrame, np.ndarray], 
                y: Union[pd.Series, np.ndarray], **kwargs) -> Dict[str, Any]:
        """
        Evaluate the model on the given data.
        
        Args:
            model: Fitted model to evaluate.
            X: Input features.
            y: Target variable.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Dictionary of evaluation metrics.
        """
        pass


class Deployer(SBYBComponent):
    """Base class for model deployment components."""
    
    @abstractmethod
    def deploy(self, model: Model, **kwargs) -> Dict[str, Any]:
        """
        Deploy the model.
        
        Args:
            model: Fitted model to deploy.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Dictionary with deployment information.
        """
        pass


class UIGenerator(SBYBComponent):
    """Base class for UI generation components."""
    
    @abstractmethod
    def generate(self, model: Model, **kwargs) -> Dict[str, Any]:
        """
        Generate a UI for the model.
        
        Args:
            model: Fitted model to create UI for.
            **kwargs: Additional keyword arguments.
            
        Returns:
            Dictionary with UI information.
        """
        pass


class Plugin(SBYBComponent):
    """Base class for plugins."""
    
    @property
    @abstractmethod
    def plugin_type(self) -> str:
        """Return the type of the plugin."""
        pass



"""
Feature selection component for SBYB AutoML.

This module provides functionality for selecting the most important features
for machine learning models.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.feature_selection import (
    SelectKBest, f_classif, f_regression, mutual_info_classif, 
    mutual_info_regression, VarianceThreshold, SelectFromModel
)
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import Lasso, LogisticRegression

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import FeatureSelectionError


class FeatureSelector(SBYBComponent):
    """
    Feature selection component.
    
    This component selects the most important features for machine learning models
    using various feature selection methods.
    """
    
    def __init__(self, method: str = 'auto', threshold: float = 0.05, 
                 max_features: Optional[int] = None, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the feature selector.
        
        Args:
            method: Feature selection method ('auto', 'variance', 'statistical', 'model_based', 'combined').
            threshold: Threshold for feature selection.
            max_features: Maximum number of features to select. If None, will be determined automatically.
            config: Configuration dictionary for the feature selector.
        """
        super().__init__(config)
        
        self.method = method
        self.threshold = threshold
        self.max_features = max_features
        self.selected_features = None
        self.feature_importance = None
    
    def select_features(self, X_train: pd.DataFrame, X_val: pd.DataFrame, y: Union[pd.Series, np.ndarray], 
                       task: str) -> Tuple[pd.DataFrame, pd.DataFrame, List[str]]:
        """
        Select features from the input data.
        
        Args:
            X_train: Training features.
            X_val: Validation features.
            y: Target variable.
            task: Machine learning task type.
            
        Returns:
            Tuple of (X_train_selected, X_val_selected, selected_feature_names).
        """
        # Choose feature selection method
        if self.method == 'auto':
            # Choose method based on data characteristics
            n_samples, n_features = X_train.shape
            
            if n_features > 100:
                # For high-dimensional data, use combined approach
                return self._combined_selection(X_train, X_val, y, task)
            elif n_features > 20:
                # For medium-dimensional data, use model-based selection
                return self._model_based_selection(X_train, X_val, y, task)
            else:
                # For low-dimensional data, use all features
                self.selected_features = X_train.columns.tolist()
                return X_train, X_val, self.selected_features
        
        elif self.method == 'variance':
            return self._variance_selection(X_train, X_val)
        
        elif self.method == 'statistical':
            return self._statistical_selection(X_train, X_val, y, task)
        
        elif self.method == 'model_based':
            return self._model_based_selection(X_train, X_val, y, task)
        
        elif self.method == 'combined':
            return self._combined_selection(X_train, X_val, y, task)
        
        else:
            raise FeatureSelectionError(f"Unknown feature selection method: {self.method}")
    
    def _variance_selection(self, X_train: pd.DataFrame, X_val: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, List[str]]:
        """
        Select features based on variance.
        
        Args:
            X_train: Training features.
            X_val: Validation features.
            
        Returns:
            Tuple of (X_train_selected, X_val_selected, selected_feature_names).
        """
        # Remove features with low variance
        selector = VarianceThreshold(threshold=self.threshold)
        X_train_array = selector.fit_transform(X_train)
        X_val_array = selector.transform(X_val)
        
        # Get selected feature names
        selected_indices = selector.get_support(indices=True)
        selected_features = [X_train.columns[i] for i in selected_indices]
        
        # Convert back to DataFrame
        X_train_selected = pd.DataFrame(X_train_array, columns=selected_features, index=X_train.index)
        X_val_selected = pd.DataFrame(X_val_array, columns=selected_features, index=X_val.index)
        
        self.selected_features = selected_features
        return X_train_selected, X_val_selected, selected_features
    
    def _statistical_selection(self, X_train: pd.DataFrame, X_val: pd.DataFrame, 
                              y: Union[pd.Series, np.ndarray], task: str) -> Tuple[pd.DataFrame, pd.DataFrame, List[str]]:
        """
        Select features based on statistical tests.
        
        Args:
            X_train: Training features.
            X_val: Validation features.
            y: Target variable.
            task: Machine learning task type.
            
        Returns:
            Tuple of (X_train_selected, X_val_selected, selected_feature_names).
        """
        # Choose statistical test based on task
        if task in ['binary', 'multiclass']:
            score_func = f_classif
        else:
            score_func = f_regression
        
        # Determine number of features to select
        k = self.max_features if self.max_features is not None else max(1, int(X_train.shape[1] * 0.8))
        
        # Select features
        selector = SelectKBest(score_func=score_func, k=k)
        X_train_array = selector.fit_transform(X_train, y)
        X_val_array = selector.transform(X_val)
        
        # Get selected feature names
        selected_indices = selector.get_support(indices=True)
        selected_features = [X_train.columns[i] for i in selected_indices]
        
        # Convert back to DataFrame
        X_train_selected = pd.DataFrame(X_train_array, columns=selected_features, index=X_train.index)
        X_val_selected = pd.DataFrame(X_val_array, columns=selected_features, index=X_val.index)
        
        # Store feature importance
        self.feature_importance = pd.DataFrame({
            'feature': X_train.columns,
            'importance': selector.scores_
        }).sort_values('importance', ascending=False)
        
        self.selected_features = selected_features
        return X_train_selected, X_val_selected, selected_features
    
    def _model_based_selection(self, X_train: pd.DataFrame, X_val: pd.DataFrame, 
                              y: Union[pd.Series, np.ndarray], task: str) -> Tuple[pd.DataFrame, pd.DataFrame, List[str]]:
        """
        Select features based on model importance.
        
        Args:
            X_train: Training features.
            X_val: Validation features.
            y: Target variable.
            task: Machine learning task type.
            
        Returns:
            Tuple of (X_train_selected, X_val_selected, selected_feature_names).
        """
        # Choose model based on task
        if task in ['binary', 'multiclass']:
            model = RandomForestClassifier(n_estimators=100, random_state=42)
        else:
            model = RandomForestRegressor(n_estimators=100, random_state=42)
        
        # Fit model
        model.fit(X_train, y)
        
        # Get feature importance
        importance = model.feature_importances_
        
        # Store feature importance
        self.feature_importance = pd.DataFrame({
            'feature': X_train.columns,
            'importance': importance
        }).sort_values('importance', ascending=False)
        
        # Select features based on importance threshold or max_features
        if self.max_features is not None:
            # Select top k features
            k = min(self.max_features, X_train.shape[1])
            threshold = sorted(importance, reverse=True)[k-1]
        else:
            # Use importance threshold
            threshold = self.threshold
        
        # Create selector
        selector = SelectFromModel(model, threshold=threshold, prefit=True)
        X_train_array = selector.transform(X_train)
        X_val_array = selector.transform(X_val)
        
        # Get selected feature names
        selected_indices = selector.get_support(indices=True)
        selected_features = [X_train.columns[i] for i in selected_indices]
        
        # Convert back to DataFrame
        X_train_selected = pd.DataFrame(X_train_array, columns=selected_features, index=X_train.index)
        X_val_selected = pd.DataFrame(X_val_array, columns=selected_features, index=X_val.index)
        
        self.selected_features = selected_features
        return X_train_selected, X_val_selected, selected_features
    
    def _combined_selection(self, X_train: pd.DataFrame, X_val: pd.DataFrame, 
                           y: Union[pd.Series, np.ndarray], task: str) -> Tuple[pd.DataFrame, pd.DataFrame, List[str]]:
        """
        Select features using a combination of methods.
        
        Args:
            X_train: Training features.
            X_val: Validation features.
            y: Target variable.
            task: Machine learning task type.
            
        Returns:
            Tuple of (X_train_selected, X_val_selected, selected_feature_names).
        """
        # Step 1: Remove low variance features
        var_selector = VarianceThreshold(threshold=0.01)
        X_train_var = pd.DataFrame(
            var_selector.fit_transform(X_train),
            columns=X_train.columns[var_selector.get_support()],
            index=X_train.index
        )
        X_val_var = pd.DataFrame(
            var_selector.transform(X_val),
            columns=X_train.columns[var_selector.get_support()],
            index=X_val.index
        )
        
        # If too few features remain, return them
        if X_train_var.shape[1] <= 5:
            self.selected_features = X_train_var.columns.tolist()
            return X_train_var, X_val_var, self.selected_features
        
        # Step 2: Apply statistical selection
        if task in ['binary', 'multiclass']:
            score_func = f_classif
        else:
            score_func = f_regression
        
        # Select a larger subset of features with statistical test
        k_stat = max(5, int(X_train_var.shape[1] * 0.5))
        stat_selector = SelectKBest(score_func=score_func, k=k_stat)
        X_train_stat = pd.DataFrame(
            stat_selector.fit_transform(X_train_var, y),
            columns=X_train_var.columns[stat_selector.get_support()],
            index=X_train_var.index
        )
        X_val_stat = pd.DataFrame(
            stat_selector.transform(X_val_var),
            columns=X_train_var.columns[stat_selector.get_support()],
            index=X_val_var.index
        )
        
        # If too few features remain, return them
        if X_train_stat.shape[1] <= 5:
            self.selected_features = X_train_stat.columns.tolist()
            return X_train_stat, X_val_stat, self.selected_features
        
        # Step 3: Apply model-based selection
        if task in ['binary', 'multiclass']:
            model = RandomForestClassifier(n_estimators=100, random_state=42)
        else:
            model = RandomForestRegressor(n_estimators=100, random_state=42)
        
        # Fit model
        model.fit(X_train_stat, y)
        
        # Get feature importance
        importance = model.feature_importances_
        
        # Store feature importance
        self.feature_importance = pd.DataFrame({
            'feature': X_train_stat.columns,
            'importance': importance
        }).sort_values('importance', ascending=False)
        
        # Determine final number of features
        if self.max_features is not None:
            k_final = min(self.max_features, X_train_stat.shape[1])
        else:
            k_final = max(5, int(X_train_stat.shape[1] * 0.8))
        
        # Select top k features
        threshold = sorted(importance, reverse=True)[min(k_final-1, len(importance)-1)]
        model_selector = SelectFromModel(model, threshold=threshold, prefit=True)
        
        X_train_final = pd.DataFrame(
            model_selector.transform(X_train_stat),
            columns=X_train_stat.columns[model_selector.get_support()],
            index=X_train_stat.index
        )
        X_val_final = pd.DataFrame(
            model_selector.transform(X_val_stat),
            columns=X_train_stat.columns[model_selector.get_support()],
            index=X_val_stat.index
        )
        
        self.selected_features = X_train_final.columns.tolist()
        return X_train_final, X_val_final, self.selected_features
    
    def get_feature_importance(self) -> Optional[pd.DataFrame]:
        """
        Get feature importance.
        
        Returns:
            DataFrame with feature names and importance scores, or None if not available.
        """
        return self.feature_importance



"""
Model monitoring component for SBYB deployment.

This module provides functionality for monitoring machine learning models
in production environments.
"""

from typing import Any, Dict, List, Optional, Union, Tuple, Callable
import os
import json
import time
import datetime
import logging
import threading
import queue
import numpy as np
import pandas as pd
from collections import deque

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import MonitoringError


class ModelMonitor(SBYBComponent):
    """
    Model monitoring component.
    
    This component monitors machine learning models in production environments.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the model monitor.
        
        Args:
            config: Configuration dictionary for the monitor.
        """
        super().__init__(config)
        self.logger = logging.getLogger("sbyb.deployment.monitoring")
        self.metrics = {}
        self.alerts = []
        self.predictions_log = deque(maxlen=10000)  # Store last 10,000 predictions
        self.monitoring_thread = None
        self.stop_monitoring = threading.Event()
        self.metrics_queue = queue.Queue()
        self.alert_callbacks = []
    
    def start_monitoring(self, model: Any, check_interval: int = 60) -> None:
        """
        Start monitoring a model.
        
        Args:
            model: The model to monitor.
            check_interval: Interval between checks in seconds.
        """
        if self.monitoring_thread is not None and self.monitoring_thread.is_alive():
            raise MonitoringError("Monitoring is already running.")
        
        # Reset stop flag
        self.stop_monitoring.clear()
        
        # Start monitoring thread
        self.monitoring_thread = threading.Thread(
            target=self._monitoring_loop,
            args=(model, check_interval),
            daemon=True
        )
        self.monitoring_thread.start()
        
        self.logger.info(f"Model monitoring started with check interval of {check_interval} seconds.")
    
    def stop_monitoring(self) -> None:
        """
        Stop monitoring.
        """
        if self.monitoring_thread is None or not self.monitoring_thread.is_alive():
            self.logger.warning("Monitoring is not running.")
            return
        
        # Set stop flag
        self.stop_monitoring.set()
        
        # Wait for thread to terminate
        self.monitoring_thread.join(timeout=5.0)
        
        if self.monitoring_thread.is_alive():
            self.logger.warning("Monitoring thread did not terminate gracefully.")
        else:
            self.logger.info("Model monitoring stopped.")
    
    def log_prediction(self, inputs: Any, prediction: Any, 
                      ground_truth: Optional[Any] = None,
                      metadata: Optional[Dict[str, Any]] = None) -> None:
        """
        Log a prediction for monitoring.
        
        Args:
            inputs: Input data for the prediction.
            prediction: Model prediction.
            ground_truth: Actual ground truth (if available).
            metadata: Additional metadata.
        """
        # Create prediction log entry
        timestamp = datetime.datetime.now().isoformat()
        
        log_entry = {
            'timestamp': timestamp,
            'prediction': self._convert_to_serializable(prediction),
            'metadata': metadata or {}
        }
        
        # Add inputs if not too large
        try:
            log_entry['inputs'] = self._convert_to_serializable(inputs)
        except Exception as e:
            log_entry['inputs'] = str(e)
        
        # Add ground truth if available
        if ground_truth is not None:
            log_entry['ground_truth'] = self._convert_to_serializable(ground_truth)
        
        # Add to predictions log
        self.predictions_log.append(log_entry)
        
        # Add to metrics queue for processing
        self.metrics_queue.put(log_entry)
    
    def add_metric(self, name: str, calculator: Callable, 
                  threshold: Optional[float] = None,
                  alert_condition: Optional[str] = None) -> None:
        """
        Add a metric to monitor.
        
        Args:
            name: Name of the metric.
            calculator: Function to calculate the metric.
            threshold: Threshold for alerting.
            alert_condition: Condition for alerting ('above', 'below', 'equal').
        """
        self.metrics[name] = {
            'calculator': calculator,
            'threshold': threshold,
            'alert_condition': alert_condition,
            'values': [],
            'timestamps': []
        }
        
        self.logger.info(f"Added metric '{name}' to monitoring.")
    
    def add_alert_callback(self, callback: Callable[[Dict[str, Any]], None]) -> None:
        """
        Add a callback function for alerts.
        
        Args:
            callback: Function to call when an alert is triggered.
        """
        self.alert_callbacks.append(callback)
    
    def get_metrics(self) -> Dict[str, Any]:
        """
        Get all metrics.
        
        Returns:
            Dictionary of metrics.
        """
        # Create a copy of metrics without the calculator functions
        metrics_copy = {}
        for name, metric in self.metrics.items():
            metrics_copy[name] = {
                'threshold': metric['threshold'],
                'alert_condition': metric['alert_condition'],
                'values': metric['values'],
                'timestamps': metric['timestamps']
            }
        
        return metrics_copy
    
    def get_alerts(self) -> List[Dict[str, Any]]:
        """
        Get all alerts.
        
        Returns:
            List of alerts.
        """
        return self.alerts
    
    def get_predictions_log(self, limit: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Get the predictions log.
        
        Args:
            limit: Maximum number of log entries to return.
            
        Returns:
            List of prediction log entries.
        """
        if limit is None or limit >= len(self.predictions_log):
            return list(self.predictions_log)
        else:
            return list(self.predictions_log)[-limit:]
    
    def export_metrics(self, output_file: str) -> None:
        """
        Export metrics to a file.
        
        Args:
            output_file: Path to the output file.
        """
        # Create output directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(output_file)), exist_ok=True)
        
        # Get metrics
        metrics = self.get_metrics()
        
        # Add export timestamp
        metrics['export_timestamp'] = datetime.datetime.now().isoformat()
        
        # Write to file
        with open(output_file, 'w') as f:
            json.dump(metrics, f, indent=2)
        
        self.logger.info(f"Metrics exported to {output_file}")
    
    def export_predictions_log(self, output_file: str) -> None:
        """
        Export predictions log to a file.
        
        Args:
            output_file: Path to the output file.
        """
        # Create output directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(output_file)), exist_ok=True)
        
        # Get predictions log
        predictions_log = self.get_predictions_log()
        
        # Write to file
        with open(output_file, 'w') as f:
            json.dump(predictions_log, f, indent=2)
        
        self.logger.info(f"Predictions log exported to {output_file}")
    
    def calculate_drift(self, reference_data: pd.DataFrame, current_data: pd.DataFrame,
                       method: str = 'ks_test') -> Dict[str, Any]:
        """
        Calculate data drift between reference and current data.
        
        Args:
            reference_data: Reference data.
            current_data: Current data.
            method: Method for calculating drift.
            
        Returns:
            Dictionary with drift metrics.
        """
        from scipy import stats
        
        if method not in ['ks_test', 'chi2_test', 'js_divergence']:
            raise MonitoringError(f"Unsupported drift detection method: {method}")
        
        # Ensure data has the same columns
        common_columns = set(reference_data.columns) & set(current_data.columns)
        if not common_columns:
            raise MonitoringError("No common columns between reference and current data.")
        
        # Calculate drift for each column
        drift_metrics = {}
        
        for column in common_columns:
            ref_values = reference_data[column].dropna()
            cur_values = current_data[column].dropna()
            
            # Skip if not enough data
            if len(ref_values) < 10 or len(cur_values) < 10:
                drift_metrics[column] = {
                    'status': 'skipped',
                    'reason': 'Not enough data'
                }
                continue
            
            # Calculate drift based on method
            if method == 'ks_test':
                # Kolmogorov-Smirnov test for continuous data
                try:
                    statistic, p_value = stats.ks_2samp(ref_values, cur_values)
                    drift_metrics[column] = {
                        'method': 'ks_test',
                        'statistic': float(statistic),
                        'p_value': float(p_value),
                        'drift_detected': p_value < 0.05
                    }
                except Exception as e:
                    drift_metrics[column] = {
                        'status': 'error',
                        'reason': str(e)
                    }
            
            elif method == 'chi2_test':
                # Chi-squared test for categorical data
                try:
                    # Get unique values
                    all_values = pd.concat([ref_values, cur_values]).unique()
                    
                    # Count occurrences
                    ref_counts = pd.Series(ref_values).value_counts().reindex(all_values).fillna(0)
                    cur_counts = pd.Series(cur_values).value_counts().reindex(all_values).fillna(0)
                    
                    # Chi-squared test
                    statistic, p_value = stats.chisquare(cur_counts, ref_counts)
                    
                    drift_metrics[column] = {
                        'method': 'chi2_test',
                        'statistic': float(statistic),
                        'p_value': float(p_value),
                        'drift_detected': p_value < 0.05
                    }
                except Exception as e:
                    drift_metrics[column] = {
                        'status': 'error',
                        'reason': str(e)
                    }
            
            elif method == 'js_divergence':
                # Jensen-Shannon divergence
                try:
                    # Get unique values
                    all_values = pd.concat([ref_values, cur_values]).unique()
                    
                    # Count occurrences
                    ref_counts = pd.Series(ref_values).value_counts().reindex(all_values).fillna(0)
                    cur_counts = pd.Series(cur_values).value_counts().reindex(all_values).fillna(0)
                    
                    # Convert to probabilities
                    ref_probs = ref_counts / ref_counts.sum()
                    cur_probs = cur_counts / cur_counts.sum()
                    
                    # Calculate JS divergence
                    m = 0.5 * (ref_probs + cur_probs)
                    js_div = 0.5 * (stats.entropy(ref_probs, m) + stats.entropy(cur_probs, m))
                    
                    drift_metrics[column] = {
                        'method': 'js_divergence',
                        'divergence': float(js_div),
                        'drift_detected': js_div > 0.1  # Threshold can be adjusted
                    }
                except Exception as e:
                    drift_metrics[column] = {
                        'status': 'error',
                        'reason': str(e)
                    }
        
        # Overall drift status
        drift_detected = any(
            col_metrics.get('drift_detected', False)
            for col_metrics in drift_metrics.values()
            if isinstance(col_metrics, dict) and 'drift_detected' in col_metrics
        )
        
        return {
            'drift_detected': drift_detected,
            'column_metrics': drift_metrics,
            'timestamp': datetime.datetime.now().isoformat()
        }
    
    def calculate_performance_metrics(self, predictions: List[Any], 
                                     ground_truths: List[Any],
                                     task_type: str) -> Dict[str, float]:
        """
        Calculate performance metrics for a model.
        
        Args:
            predictions: List of predictions.
            ground_truths: List of ground truths.
            task_type: Type of task ('classification', 'regression').
            
        Returns:
            Dictionary of performance metrics.
        """
        if task_type == 'classification':
            from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
            
            try:
                metrics = {
                    'accuracy': float(accuracy_score(ground_truths, predictions)),
                    'precision': float(precision_score(ground_truths, predictions, average='weighted')),
                    'recall': float(recall_score(ground_truths, predictions, average='weighted')),
                    'f1': float(f1_score(ground_truths, predictions, average='weighted'))
                }
            except Exception as e:
                raise MonitoringError(f"Error calculating classification metrics: {str(e)}")
        
        elif task_type == 'regression':
            from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
            
            try:
                mse = mean_squared_error(ground_truths, predictions)
                metrics = {
                    'mse': float(mse),
                    'rmse': float(np.sqrt(mse)),
                    'mae': float(mean_absolute_error(ground_truths, predictions)),
                    'r2': float(r2_score(ground_truths, predictions))
                }
            except Exception as e:
                raise MonitoringError(f"Error calculating regression metrics: {str(e)}")
        
        else:
            raise MonitoringError(f"Unsupported task type: {task_type}")
        
        return metrics
    
    def _monitoring_loop(self, model: Any, check_interval: int) -> None:
        """
        Main monitoring loop.
        
        Args:
            model: The model to monitor.
            check_interval: Interval between checks in seconds.
        """
        while not self.stop_monitoring.is_set():
            try:
                # Process metrics queue
                self._process_metrics_queue()
                
                # Check for alerts
                self._check_alerts()
                
                # Wait for next check
                self.stop_monitoring.wait(check_interval)
            except Exception as e:
                self.logger.error(f"Error in monitoring loop: {str(e)}")
    
    def _process_metrics_queue(self) -> None:
        """
        Process metrics queue.
        """
        # Get all items from queue
        items = []
        while not self.metrics_queue.empty():
            try:
                items.append(self.metrics_queue.get_nowait())
                self.metrics_queue.task_done()
            except queue.Empty:
                break
        
        if not items:
            return
        
        # Calculate metrics
        timestamp = datetime.datetime.now().isoformat()
        
        for name, metric in self.metrics.items():
            calculator = metric['calculator']
            
            try:
                # Calculate metric
                value = calculator(items)
                
                # Add to metric values
                metric['values'].append(value)
                metric['timestamps'].append(timestamp)
                
                # Check for alert
                if metric['threshold'] is not None and metric['alert_condition'] is not None:
                    threshold = metric['threshold']
                    condition = metric['alert_condition']
                    
                    alert_triggered = False
                    
                    if condition == 'above' and value > threshold:
                        alert_triggered = True
                    elif condition == 'below' and value < threshold:
                        alert_triggered = True
                    elif condition == 'equal' and value == threshold:
                        alert_triggered = True
                    
                    if alert_triggered:
                        alert = {
                            'timestamp': timestamp,
                            'metric': name,
                            'value': value,
                            'threshold': threshold,
                            'condition': condition
                        }
                        
                        self.alerts.append(alert)
                        self.logger.warning(f"Alert triggered for metric '{name}': {value} {condition} {threshold}")
                        
                        # Call alert callbacks
                        for callback in self.alert_callbacks:
                            try:
                                callback(alert)
                            except Exception as e:
                                self.logger.error(f"Error in alert callback: {str(e)}")
            
            except Exception as e:
                self.logger.error(f"Error calculating metric '{name}': {str(e)}")
    
    def _check_alerts(self) -> None:
        """
        Check for alerts.
        """
        # This method can be extended to implement more complex alert logic
        pass
    
    def _convert_to_serializable(self, obj: Any) -> Any:
        """
        Convert an object to a JSON-serializable format.
        
        Args:
            obj: Object to convert.
            
        Returns:
            JSON-serializable object.
        """
        if isinstance(obj, (str, int, float, bool, type(None))):
            return obj
        elif isinstance(obj, (list, tuple)):
            return [self._convert_to_serializable(item) for item in obj]
        elif isinstance(obj, dict):
            return {str(k): self._convert_to_serializable(v) for k, v in obj.items()}
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (pd.DataFrame, pd.Series)):
            return obj.to_dict()
        elif hasattr(obj, 'tolist'):
            return obj.tolist()
        else:
            return str(obj)



"""
Main API interface for SBYB.

This module provides the main programmatic interface for the SBYB library,
allowing users to interact with all components through a unified API.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import logging
import pandas as pd
import numpy as np
import datetime

from sbyb.core.config import Config
from sbyb.preprocessing import PreprocessingPipeline
from sbyb.task_detection import TaskDetector
from sbyb.automl import AutoMLEngine
from sbyb.evaluation import Evaluator, Explainer
from sbyb.deployment import ModelExporter, ModelServer
from sbyb.ui_generator import DashboardGenerator, FormGenerator
from sbyb.scaffolding import ProjectGenerator
from sbyb.eda import DataProfiler, Visualizer, DataAnalyzer
from sbyb.plugins import PluginManager
from sbyb.tracking import ExperimentTracker, Run, Experiment, TrackingVisualizer


class SBYB:
    """
    Main interface for the SBYB library.
    
    This class provides a unified API for interacting with all components
    of the SBYB library, making it easy to use the library programmatically.
    """
    
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the SBYB API.
        
        Args:
            config: Optional configuration dictionary.
        """
        self.config = config or {}
        
        # Set up logging
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger('sbyb.api')
        
        # Initialize components
        self._init_components()
    
    def _init_components(self):
        """
        Initialize all components of the SBYB library.
        """
        # Core components
        self.preprocessing = PreprocessingPipeline(self.config.get('preprocessing'))
        self.task_detector = TaskDetector(self.config.get('task_detection'))
        self.automl = AutoMLEngine(self.config.get('automl'))
        self.evaluator = Evaluator(self.config.get('evaluation'))
        self.explainer = Explainer(self.config.get('explainer'))
        self.model_exporter = ModelExporter(self.config.get('model_exporter'))
        self.model_server = ModelServer(self.config.get('model_server'))
        self.dashboard_generator = DashboardGenerator(self.config.get('dashboard_generator'))
        self.form_generator = FormGenerator(self.config.get('form_generator'))
        self.project_generator = ProjectGenerator(self.config.get('project_generator'))
        self.data_profiler = DataProfiler(self.config.get('data_profiler'))
        self.visualizer = Visualizer(self.config.get('visualizer'))
        self.data_analyzer = DataAnalyzer(self.config.get('data_analyzer'))
        self.plugin_manager = PluginManager(self.config.get('plugin_manager'))
        self.experiment_tracker = ExperimentTracker(self.config.get('experiment_tracker'))
        self.tracking_visualizer = TrackingVisualizer(self.config.get('tracking_visualizer'))
    
    #
    # Project Management
    #
    
    def create_project(self, name: str, template: str = 'basic', output_dir: str = '.',
                      description: Optional[str] = None) -> str:
        """
        Create a new project.
        
        Args:
            name: Project name.
            template: Project template. Options: 'basic', 'classification', 'regression', 'time_series'.
            output_dir: Output directory.
            description: Optional project description.
            
        Returns:
            Path to the created project.
        """
        self.logger.info(f"Creating new project: {name}")
        
        output_dir = os.path.abspath(output_dir)
        project_dir = os.path.join(output_dir, name)
        
        self.project_generator.create_project(
            name=name,
            template=template,
            output_dir=output_dir,
            description=description or f"{name} - A SBYB project"
        )
        
        self.logger.info(f"Project created successfully: {project_dir}")
        return project_dir
    
    #
    # Data Preprocessing
    #
    
    def preprocess_data(self, data: Union[pd.DataFrame, str], config: Optional[Dict[str, Any]] = None,
                       output_file: Optional[str] = None, save_pipeline: Optional[str] = None) -> pd.DataFrame:
        """
        Preprocess data.
        
        Args:
            data: Input data as DataFrame or path to data file.
            config: Optional preprocessing configuration.
            output_file: Optional path to save preprocessed data.
            save_pipeline: Optional path to save preprocessing pipeline.
            
        Returns:
            Preprocessed data as DataFrame.
        """
        self.logger.info("Preprocessing data")
        
        # Load data if it's a file path
        if isinstance(data, str):
            if data.endswith('.csv'):
                data = pd.read_csv(data)
            elif data.endswith(('.xls', '.xlsx')):
                data = pd.read_excel(data)
            elif data.endswith('.parquet'):
                data = pd.read_parquet(data)
            elif data.endswith('.json'):
                data = pd.read_json(data)
            else:
                raise ValueError(f"Unsupported data file format: {data}")
        
        # Update pipeline config if provided
        if config:
            self.preprocessing.update_config(config)
        
        # Preprocess data
        preprocessed_data = self.preprocessing.fit_transform(data)
        
        # Save preprocessed data
        if output_file:
            if output_file.endswith('.csv'):
                preprocessed_data.to_csv(output_file, index=False)
            elif output_file.endswith(('.xls', '.xlsx')):
                preprocessed_data.to_excel(output_file, index=False)
            elif output_file.endswith('.parquet'):
                preprocessed_data.to_parquet(output_file, index=False)
            else:
                preprocessed_data.to_csv(output_file, index=False)
            
            self.logger.info(f"Preprocessed data saved to: {output_file}")
        
        # Save pipeline
        if save_pipeline:
            self.preprocessing.save(save_pipeline)
            self.logger.info(f"Preprocessing pipeline saved to: {save_pipeline}")
        
        self.logger.info("Data preprocessing completed successfully")
        return preprocessed_data
    
    def profile_data(self, data: Union[pd.DataFrame, str], output_dir: Optional[str] = None,
                    format: str = 'html') -> Any:
        """
        Generate a data profile.
        
        Args:
            data: Input data as DataFrame or path to data file.
            output_dir: Optional directory to save profile report.
            format: Output format. Options: 'html', 'json'.
            
        Returns:
            Data profile object.
        """
        self.logger.info("Generating data profile")
        
        # Load data if it's a file path
        if isinstance(data, str):
            if data.endswith('.csv'):
                data = pd.read_csv(data)
            elif data.endswith(('.xls', '.xlsx')):
                data = pd.read_excel(data)
            elif data.endswith('.parquet'):
                data = pd.read_parquet(data)
            elif data.endswith('.json'):
                data = pd.read_json(data)
            else:
                raise ValueError(f"Unsupported data file format: {data}")
        
        # Generate profile
        profile = self.data_profiler.generate_profile(data)
        
        # Save profile
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)
            
            if format == 'html':
                output_file = os.path.join(output_dir, 'data_profile.html')
                profile.to_html(output_file)
            elif format == 'json':
                output_file = os.path.join(output_dir, 'data_profile.json')
                profile.to_json(output_file)
            
            self.logger.info(f"Data profile saved to: {output_file}")
        
        self.logger.info("Data profiling completed successfully")
        return profile
    
    #
    # AutoML
    #
    
    def run_automl(self, data: Union[pd.DataFrame, str], target: str, task: Optional[str] = None,
                  config: Optional[Dict[str, Any]] = None, output_dir: Optional[str] = None,
                  time_limit: Optional[int] = None, track: bool = False,
                  experiment_name: Optional[str] = None) -> Any:
        """
        Run AutoML.
        
        Args:
            data: Input data as DataFrame or path to data file.
            target: Target column name.
            task: Optional ML task type (auto-detected if not specified).
            config: Optional AutoML configuration.
            output_dir: Optional directory to save results.
            time_limit: Optional time limit in seconds.
            track: Whether to track the experiment.
            experiment_name: Optional experiment name for tracking.
            
        Returns:
            AutoML result object.
        """
        self.logger.info("Running AutoML")
        
        # Load data if it's a file path
        if isinstance(data, str):
            if data.endswith('.csv'):
                data = pd.read_csv(data)
            elif data.endswith(('.xls', '.xlsx')):
                data = pd.read_excel(data)
            elif data.endswith('.parquet'):
                data = pd.read_parquet(data)
            elif data.endswith('.json'):
                data = pd.read_json(data)
            else:
                raise ValueError(f"Unsupported data file format: {data}")
        
        # Split features and target
        if target not in data.columns:
            raise ValueError(f"Target column not found: {target}")
        
        X = data.drop(columns=[target])
        y = data[target]
        
        # Update AutoML config if provided
        if config:
            self.automl.update_config(config)
        
        # Set time limit if specified
        if time_limit:
            self.automl.config['time_limit'] = time_limit
        
        # Create experiment tracker if tracking is enabled
        tracker = None
        if track:
            tracker = self.experiment_tracker
            
            # Create or set experiment
            if experiment_name:
                try:
                    # Try to find existing experiment
                    experiments = tracker.list_experiments()
                    experiment_id = None
                    
                    for exp in experiments:
                        if exp['name'] == experiment_name:
                            experiment_id = exp['experiment_id']
                            break
                    
                    if experiment_id:
                        tracker.set_experiment(experiment_id)
                        self.logger.info(f"Using existing experiment: {experiment_name}")
                    else:
                        tracker.create_experiment(name=experiment_name)
                        self.logger.info(f"Created new experiment: {experiment_name}")
                except Exception as e:
                    self.logger.warning(f"Error setting experiment: {str(e)}")
                    tracker.create_experiment(name=f"AutoML Run {datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}")
            else:
                tracker.create_experiment(name=f"AutoML Run {datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}")
            
            # Create run
            run = tracker.create_run(name=f"AutoML Run {datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}")
            tracker.start_run()
            
            # Log parameters
            tracker.log_parameters({
                'target_column': target,
                'n_samples': len(data),
                'n_features': len(X.columns),
                'time_limit': time_limit
            })
        
        try:
            # Detect task type if not specified
            task_type = task
            if not task_type:
                self.logger.info("Auto-detecting task type...")
                task_type = self.task_detector.detect(X, y)
                self.logger.info(f"Detected task type: {task_type}")
                
                if tracker:
                    tracker.log_parameter('task_type', task_type)
            
            # Set task type
            self.automl.set_task(task_type)
            
            # Run AutoML
            self.logger.info("Running AutoML...")
            result = self.automl.fit(X, y)
            
            # Log metrics if tracking
            if tracker:
                tracker.log_metrics(result.metrics)
                
                # Log model
                if output_dir:
                    model_path = os.path.join(output_dir, 'model.pkl')
                    os.makedirs(output_dir, exist_ok=True)
                    result.save_model(model_path)
                    tracker.log_model(result.model, model_path)
                
                # End run
                tracker.end_run()
            
            # Save results
            if output_dir:
                os.makedirs(output_dir, exist_ok=True)
                
                # Save model
                model_path = os.path.join(output_dir, 'model.pkl')
                result.save_model(model_path)
                self.logger.info(f"Model saved to: {model_path}")
                
                # Save leaderboard
                leaderboard_path = os.path.join(output_dir, 'leaderboard.csv')
                result.leaderboard.to_csv(leaderboard_path, index=False)
                self.logger.info(f"Leaderboard saved to: {leaderboard_path}")
                
                # Save feature importance
                if hasattr(result, 'feature_importance') and result.feature_importance is not None:
                    fi_path = os.path.join(output_dir, 'feature_importance.csv')
                    result.feature_importance.to_csv(fi_path, index=False)
                    self.logger.info(f"Feature importance saved to: {fi_path}")
                
                # Save predictions
                preds_path = os.path.join(output_dir, 'predictions.csv')
                result.predictions.to_csv(preds_path, index=False)
                self.logger.info(f"Predictions saved to: {preds_path}")
            
            self.logger.info("AutoML completed successfully")
            return result
        
        except Exception as e:
            self.logger.error(f"Error running AutoML: {str(e)}")
            
            # End run with failure if tracking
            if tracker and tracker.current_run:
                tracker.end_run(status="failed", error_message=str(e))
            
            raise
    
    #
    # Evaluation
    #
    
    def evaluate_model(self, model: Any, data: Union[pd.DataFrame, str], target: str,
                      metrics: Optional[List[str]] = None, output_dir: Optional[str] = None) -> Any:
        """
        Evaluate a model.
        
        Args:
            model: Model object or path to model file.
            data: Test data as DataFrame or path to data file.
            target: Target column name.
            metrics: Optional list of metrics to compute.
            output_dir: Optional directory to save evaluation results.
            
        Returns:
            Evaluation result object.
        """
        self.logger.info("Evaluating model")
        
        # Load model if it's a file path
        if isinstance(model, str):
            import pickle
            with open(model, 'rb') as f:
                model = pickle.load(f)
        
        # Load data if it's a file path
        if isinstance(data, str):
            if data.endswith('.csv'):
                data = pd.read_csv(data)
            elif data.endswith(('.xls', '.xlsx')):
                data = pd.read_excel(data)
            elif data.endswith('.parquet'):
                data = pd.read_parquet(data)
            elif data.endswith('.json'):
                data = pd.read_json(data)
            else:
                raise ValueError(f"Unsupported data file format: {data}")
        
        # Split features and target
        if target not in data.columns:
            raise ValueError(f"Target column not found: {target}")
        
        X = data.drop(columns=[target])
        y = data[target]
        
        # Evaluate model
        evaluation = self.evaluator.evaluate(model, X, y, metrics=metrics)
        
        # Save results
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)
            
            # Save metrics
            metrics_path = os.path.join(output_dir, 'metrics.json')
            import json
            with open(metrics_path, 'w') as f:
                json.dump(evaluation.metrics, f, indent=2)
            self.logger.info(f"Metrics saved to: {metrics_path}")
            
            # Save confusion matrix if available
            if hasattr(evaluation, 'confusion_matrix') and evaluation.confusion_matrix is not None:
                cm_path = os.path.join(output_dir, 'confusion_matrix.csv')
                evaluation.confusion_matrix.to_csv(cm_path, index=False)
                self.logger.info(f"Confusion matrix saved to: {cm_path}")
            
            # Save ROC curve if available
            if hasattr(evaluation, 'roc_curve') and evaluation.roc_curve is not None:
                roc_path = os.path.join(output_dir, 'roc_curve.png')
                evaluation.roc_curve.savefig(roc_path)
                self.logger.info(f"ROC curve saved to: {roc_path}")
            
            # Save predictions
            preds_path = os.path.join(output_dir, 'predictions.csv')
            evaluation.predictions.to_csv(preds_path, index=False)
            self.logger.info(f"Predictions saved to: {preds_path}")
        
        self.logger.info("Evaluation completed successfully")
        return evaluation
    
    def explain_model(self, model: Any, data: Union[pd.DataFrame, str], method: str = 'shap',
                     output_dir: Optional[str] = None) -> Any:
        """
        Explain a model.
        
        Args:
            model: Model object or path to model file.
            data: Data as DataFrame or path to data file.
            method: Explanation method. Options: 'shap', 'lime', 'eli5', 'all'.
            output_dir: Optional directory to save explanation results.
            
        Returns:
            Explanation result object.
        """
        self.logger.info(f"Explaining model using {method}")
        
        # Load model if it's a file path
        if isinstance(model, str):
            import pickle
            with open(model, 'rb') as f:
                model = pickle.load(f)
        
        # Load data if it's a file path
        if isinstance(data, str):
            if data.endswith('.csv'):
                data = pd.read_csv(data)
            elif data.endswith(('.xls', '.xlsx')):
                data = pd.read_excel(data)
            elif data.endswith('.parquet'):
                data = pd.read_parquet(data)
            elif data.endswith('.json'):
                data = pd.read_json(data)
            else:
                raise ValueError(f"Unsupported data file format: {data}")
        
        # Generate explanations
        explanation = self.explainer.explain(model, data, method=method)
        
        # Save results
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)
            
            # Save feature importance
            if hasattr(explanation, 'feature_importance') and explanation.feature_importance is not None:
                fi_path = os.path.join(output_dir, 'feature_importance.csv')
                explanation.feature_importance.to_csv(fi_path, index=False)
                self.logger.info(f"Feature importance saved to: {fi_path}")
            
            # Save SHAP values if available
            if hasattr(explanation, 'shap_values') and explanation.shap_values is not None:
                shap_path = os.path.join(output_dir, 'shap_values.csv')
                pd.DataFrame(explanation.shap_values).to_csv(shap_path, index=False)
                self.logger.info(f"SHAP values saved to: {shap_path}")
            
            # Save SHAP summary plot if available
            if hasattr(explanation, 'shap_summary_plot') and explanation.shap_summary_plot is not None:
                shap_plot_path = os.path.join(output_dir, 'shap_summary.png')
                explanation.shap_summary_plot.savefig(shap_plot_path)
                self.logger.info(f"SHAP summary plot saved to: {shap_plot_path}")
            
            # Save explanation report
            report_path = os.path.join(output_dir, 'explanation_report.html')
            explanation.to_html(report_path)
            self.logger.info(f"Explanation report saved to: {report_path}")
        
        self.logger.info("Explanation completed successfully")
        return explanation
    
    #
    # Deployment
    #
    
    def export_model(self, model: Any, format: str = 'pickle', output: Optional[str] = None) -> str:
        """
        Export a model.
        
        Args:
            model: Model object or path to model file.
            format: Export format. Options: 'pickle', 'onnx', 'tensorflow', 'mlflow'.
            output: Optional output file or directory.
            
        Returns:
            Path to the exported model.
        """
        self.logger.info(f"Exporting model to {format} format")
        
        # Load model if it's a file path
        if isinstance(model, str):
            import pickle
            with open(model, 'rb') as f:
                model = pickle.load(f)
        
        # Export model
        if format == 'pickle':
            output_file = output or 'model_exported.pkl'
            path = self.model_exporter.to_pickle(model, output_file)
        elif format == 'onnx':
            output_file = output or 'model.onnx'
            path = self.model_exporter.to_onnx(model, output_file)
        elif format == 'tensorflow':
            output_dir = output or 'model_tf'
            path = self.model_exporter.to_tensorflow(model, output_dir)
        elif format == 'mlflow':
            output_dir = output or 'model_mlflow'
            path = self.model_exporter.to_mlflow(model, output_dir)
        else:
            raise ValueError(f"Unsupported export format: {format}")
        
        self.logger.info(f"Model exported to: {path}")
        return path
    
    def serve_model(self, model: Any, host: str = '0.0.0.0', port: int = 8000,
                   server_type: str = 'fastapi') -> Any:
        """
        Serve a model.
        
        Args:
            model: Model object or path to model file.
            host: Host to bind to.
            port: Port to bind to.
            server_type: Server type. Options: 'fastapi', 'flask', 'streamlit'.
            
        Returns:
            Server object.
        """
        self.logger.info(f"Serving model with {server_type} server on {host}:{port}")
        
        # Load model if it's a file path
        if isinstance(model, str):
            import pickle
            with open(model, 'rb') as f:
                model = pickle.load(f)
        
        # Set server type
        self.model_server.set_server_type(server_type)
        
        # Serve model
        server = self.model_server.serve(model, host=host, port=port, start=False)
        
        self.logger.info(f"Model server created. Call start() on the returned object to start the server.")
        return server
    
    #
    # UI Generation
    #
    
    def generate_ui(self, model: Any, output_dir: str, ui_type: str = 'dashboard',
                   framework: str = 'streamlit', theme: Optional[str] = None) -> str:
        """
        Generate a UI.
        
        Args:
            model: Model object or path to model file.
            output_dir: Output directory.
            ui_type: UI type. Options: 'dashboard', 'form', 'app'.
            framework: UI framework. Options: 'streamlit', 'dash', 'flask', 'html'.
            theme: Optional UI theme.
            
        Returns:
            Path to the generated UI.
        """
        self.logger.info(f"Generating {ui_type} UI with {framework} framework")
        
        # Load model if it's a file path
        if isinstance(model, str):
            import pickle
            with open(model, 'rb') as f:
                model = pickle.load(f)
        
        # Create UI generator
        if ui_type == 'dashboard':
            generator = self.dashboard_generator
            generator.set_framework(framework)
        elif ui_type == 'form':
            generator = self.form_generator
            generator.set_framework(framework)
        else:  # app
            if framework == 'streamlit':
                from sbyb.ui_generator import StreamlitAppGenerator
                generator = StreamlitAppGenerator()
            elif framework == 'dash':
                from sbyb.ui_generator import DashAppGenerator
                generator = DashAppGenerator()
            elif framework == 'flask':
                from sbyb.ui_generator import FlaskAppGenerator
                generator = FlaskAppGenerator()
            else:  # html
                from sbyb.ui_generator import HTMLAppGenerator
                generator = HTMLAppGenerator()
        
        # Set theme if specified
        if theme:
            generator.set_theme(theme)
        
        # Generate UI
        ui_path = generator.generate(model, output_dir=output_dir)
        
        self.logger.info(f"UI generated successfully: {ui_path}")
        return ui_path
    
    #
    # Plugin Management
    #
    
    def list_plugins(self, category: Optional[str] = None) -> Dict[str, List[Dict[str, Any]]]:
        """
        List installed plugins.
        
        Args:
            category: Optional category to filter by.
            
        Returns:
            Dictionary of plugin categories and their available plugins with metadata.
        """
        self.logger.info("Listing installed plugins")
        return self.plugin_manager.list_plugins(category=category)
    
    def install_plugin(self, source: str, force: bool = False) -> bool:
        """
        Install a plugin.
        
        Args:
            source: Plugin source (local path, git URL, or pip package).
            force: Whether to force installation if the plugin already exists.
            
        Returns:
            True if the plugin was installed successfully, False otherwise.
        """
        self.logger.info(f"Installing plugin from: {source}")
        return self.plugin_manager.install_plugin(source, force=force)
    
    def uninstall_plugin(self, name: str, category: Optional[str] = None) -> bool:
        """
        Uninstall a plugin.
        
        Args:
            name: Plugin name.
            category: Optional plugin category.
            
        Returns:
            True if the plugin was uninstalled successfully, False otherwise.
        """
        self.logger.info(f"Uninstalling plugin: {name}")
        return self.plugin_manager.uninstall_plugin(name, category=category)
    
    def create_plugin_template(self, name: str, output_dir: str = '.', category: str = 'custom',
                              description: Optional[str] = None, author: Optional[str] = None) -> bool:
        """
        Create a plugin template.
        
        Args:
            name: Plugin name.
            output_dir: Output directory.
            category: Plugin category.
            description: Optional plugin description.
            author: Optional plugin author.
            
        Returns:
            True if the template was created successfully, False otherwise.
        """
        self.logger.info(f"Creating plugin template: {name}")
        return self.plugin_manager.create_plugin_template(
            output_dir=output_dir,
            name=name,
            category=category,
            description=description or f"{name} plugin",
            author=author or "SBYB User"
        )
    
    #
    # Experiment Tracking
    #
    
    def create_experiment(self, name: str, description: Optional[str] = None,
                         tags: Optional[List[str]] = None) -> Experiment:
        """
        Create a new experiment.
        
        Args:
            name: Experiment name.
            description: Optional experiment description.
            tags: Optional list of tags.
            
        Returns:
            Created Experiment object.
        """
        self.logger.info(f"Creating experiment: {name}")
        return self.experiment_tracker.create_experiment(
            name=name,
            description=description,
            tags=tags
        )
    
    def get_experiment(self, experiment_id: str) -> Experiment:
        """
        Get an experiment by ID.
        
        Args:
            experiment_id: Experiment ID.
            
        Returns:
            Retrieved Experiment object.
        """
        return self.experiment_tracker.get_experiment(experiment_id)
    
    def list_experiments(self) -> List[Dict[str, Any]]:
        """
        List all experiments.
        
        Returns:
            List of experiment metadata dictionaries.
        """
        self.logger.info("Listing experiments")
        return self.experiment_tracker.list_experiments()
    
    def delete_experiment(self, experiment_id: str) -> bool:
        """
        Delete an experiment.
        
        Args:
            experiment_id: Experiment ID.
            
        Returns:
            True if the experiment was deleted, False otherwise.
        """
        self.logger.info(f"Deleting experiment: {experiment_id}")
        return self.experiment_tracker.delete_experiment(experiment_id)
    
    def create_run(self, experiment_id: str, name: Optional[str] = None,
                  description: Optional[str] = None, tags: Optional[List[str]] = None) -> Run:
        """
        Create a new run.
        
        Args:
            experiment_id: Experiment ID.
            name: Optional run name.
            description: Optional run description.
            tags: Optional list of tags.
            
        Returns:
            Created Run object.
        """
        self.logger.info(f"Creating run for experiment: {experiment_id}")
        
        # Set experiment
        self.experiment_tracker.set_experiment(experiment_id)
        
        return self.experiment_tracker.create_run(
            name=name,
            description=description,
            tags=tags
        )
    
    def start_run(self) -> Run:
        """
        Start the current run.
        
        Returns:
            Started Run object.
        """
        self.logger.info("Starting run")
        return self.experiment_tracker.start_run()
    
    def end_run(self, status: str = "completed", error_message: Optional[str] = None) -> Run:
        """
        End the current run.
        
        Args:
            status: Run status. Options: 'completed', 'failed'.
            error_message: Optional error message if the run failed.
            
        Returns:
            Ended Run object.
        """
        self.logger.info(f"Ending run with status: {status}")
        return self.experiment_tracker.end_run(status=status, error_message=error_message)
    
    def log_metric(self, key: str, value: Union[int, float]) -> None:
        """
        Log a metric for the current run.
        
        Args:
            key: Metric name.
            value: Metric value.
        """
        self.experiment_tracker.log_metric(key, value)
    
    def log_parameter(self, key: str, value: Any) -> None:
        """
        Log a parameter for the current run.
        
        Args:
            key: Parameter name.
            value: Parameter value.
        """
        self.experiment_tracker.log_parameter(key, value)
    
    def log_model(self, model: Any, model_path: Optional[str] = None) -> str:
        """
        Log a model for the current run.
        
        Args:
            model: Model object.
            model_path: Optional path to save the model.
            
        Returns:
            Path to the saved model.
        """
        return self.experiment_tracker.log_model(model, model_path)
    
    def visualize_experiment(self, experiment_id: str, metrics: List[str],
                            params: Optional[List[str]] = None, output_dir: Optional[str] = None,
                            format: str = 'html') -> str:
        """
        Visualize an experiment.
        
        Args:
            experiment_id: Experiment ID.
            metrics: List of metrics to visualize.
            params: Optional list of parameters to visualize.
            output_dir: Optional directory to save visualizations.
            format: Output format. Options: 'html', 'png', 'pdf'.
            
        Returns:
            Path to the visualization report.
        """
        self.logger.info(f"Visualizing experiment: {experiment_id}")
        
        # Create output directory if specified
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)
        
        # Generate visualizations
        if format == 'html':
            # Generate HTML report
            report_path = self.tracking_visualizer.export_experiment_report(
                experiment_id=experiment_id,
                output_dir=output_dir or os.getcwd(),
                metrics=metrics,
                params=params
            )
            
            self.logger.info(f"Visualization report saved to: {report_path}")
            return report_path
        else:
            # Generate individual visualizations
            experiment = self.experiment_tracker.get_experiment(experiment_id)
            
            # Get run IDs
            run_ids = [run.run_id for run in experiment.runs]
            
            # Generate metric comparison plots
            for metric in metrics:
                try:
                    fig = self.tracking_visualizer.plot_metric_comparison(
                        run_ids=run_ids,
                        metric=metric,
                        use_plotly=False
                    )
                    
                    if output_dir:
                        output_file = os.path.join(output_dir, f"metric_{metric}.{format}")
                        fig.savefig(output_file)
                        self.logger.info(f"Metric comparison plot saved to: {output_file}")
                except Exception as e:
                    self.logger.warning(f"Error generating metric comparison for {metric}: {str(e)}")
            
            # Generate parameter importance plots
            if params:
                for metric in metrics:
                    try:
                        fig = self.tracking_visualizer.plot_parameter_importance(
                            experiment_id=experiment_id,
                            metric=metric,
                            use_plotly=False
                        )
                        
                        if output_dir:
                            output_file = os.path.join(output_dir, f"param_importance_{metric}.{format}")
                            fig.savefig(output_file)
                            self.logger.info(f"Parameter importance plot saved to: {output_file}")
                    except Exception as e:
                        self.logger.warning(f"Error generating parameter importance for {metric}: {str(e)}")
            
            # Return output directory
            return output_dir or os.getcwd()
    
    #
    # Context Manager Support
    #
    
    def __enter__(self):
        """
        Enter context manager.
        
        Returns:
            Self.
        """
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit context manager.
        
        Args:
            exc_type: Exception type.
            exc_val: Exception value.
            exc_tb: Exception traceback.
        """
        # Clean up resources if needed
        pass



"""
Experiment tracker for SBYB.

This module provides the ExperimentTracker class, which is the main interface
for tracking machine learning experiments locally without requiring external services.
"""

from typing import Any, Dict, List, Optional, Union, Tuple
import os
import json
import datetime
import uuid
from pathlib import Path
import logging

from sbyb.core.base import SBYBComponent
from sbyb.core.exceptions import TrackingError
from sbyb.tracking.experiment import Experiment
from sbyb.tracking.run import Run
from sbyb.tracking.storage import LocalStorage


class ExperimentTracker(SBYBComponent):
    """
    Main interface for tracking machine learning experiments.
    
    This component provides a unified interface for creating and managing
    experiments and runs, logging metrics, parameters, and artifacts,
    and querying tracking data.
    """
    
    def __init__(self, tracking_dir: Optional[str] = None, config: Optional[Dict[str, Any]] = None):
        """
        Initialize the experiment tracker.
        
        Args:
            tracking_dir: Directory for storing tracking data. Defaults to ~/.sbyb/tracking.
            config: Optional configuration dictionary.
        """
        super().__init__(config)
        
        # Set up logging
        self.logger = logging.getLogger("sbyb.tracking")
        
        # Initialize storage
        self.storage = LocalStorage(root_dir=tracking_dir, config=config)
        
        # Current experiment and run
        self.current_experiment = None
        self.current_run = None
    
    def create_experiment(self, name: str, description: Optional[str] = None,
                         tags: Optional[List[str]] = None) -> Experiment:
        """
        Create a new experiment.
        
        Args:
            name: Name of the experiment.
            description: Optional description of the experiment.
            tags: Optional list of tags for categorizing the experiment.
            
        Returns:
            Created Experiment object.
        """
        experiment = Experiment(name=name, description=description, tags=tags)
        self.current_experiment = experiment
        self.storage.save_experiment(experiment)
        
        self.logger.info(f"Created experiment: {name} (ID: {experiment.experiment_id})")
        return experiment
    
    def get_experiment(self, experiment_id: str) -> Experiment:
        """
        Get an experiment by ID.
        
        Args:
            experiment_id: ID of the experiment to retrieve.
            
        Returns:
            Retrieved Experiment object.
            
        Raises:
            TrackingError: If the experiment is not found.
        """
        return self.storage.load_experiment(experiment_id)
    
    def set_experiment(self, experiment_id: str) -> Experiment:
        """
        Set the current experiment.
        
        Args:
            experiment_id: ID of the experiment to set as current.
            
        Returns:
            Set Experiment object.
            
        Raises:
            TrackingError: If the experiment is not found.
        """
        experiment = self.storage.load_experiment(experiment_id)
        self.current_experiment = experiment
        
        self.logger.info(f"Set current experiment: {experiment.name} (ID: {experiment.experiment_id})")
        return experiment
    
    def list_experiments(self) -> List[Dict[str, Any]]:
        """
        List all experiments.
        
        Returns:
            List of experiment metadata dictionaries.
        """
        return self.storage.list_experiments()
    
    def delete_experiment(self, experiment_id: str) -> bool:
        """
        Delete an experiment.
        
        Args:
            experiment_id: ID of the experiment to delete.
            
        Returns:
            True if the experiment was deleted, False otherwise.
        """
        if self.current_experiment and self.current_experiment.experiment_id == experiment_id:
            self.current_experiment = None
        
        return self.storage.delete_experiment(experiment_id)
    
    def create_run(self, name: Optional[str] = None, description: Optional[str] = None,
                  tags: Optional[List[str]] = None) -> Run:
        """
        Create a new run.
        
        Args:
            name: Optional name of the run.
            description: Optional description of the run.
            tags: Optional list of tags for categorizing the run.
            
        Returns:
            Created Run object.
            
        Raises:
            TrackingError: If no current experiment is set.
        """
        if not self.current_experiment:
            raise TrackingError("No current experiment set. Call create_experiment() or set_experiment() first.")
        
        run = Run(name=name, description=description, tags=tags)
        run.experiment_id = self.current_experiment.experiment_id
        
        self.current_experiment.add_run(run)
        self.current_run = run
        
        # Save both the run and the updated experiment
        self.storage.save_run(run)
        self.storage.save_experiment(self.current_experiment)
        
        self.logger.info(f"Created run: {run.name} (ID: {run.run_id})")
        return run
    
    def get_run(self, run_id: str) -> Run:
        """
        Get a run by ID.
        
        Args:
            run_id: ID of the run to retrieve.
            
        Returns:
            Retrieved Run object.
            
        Raises:
            TrackingError: If the run is not found.
        """
        return self.storage.load_run(run_id)
    
    def set_run(self, run_id: str) -> Run:
        """
        Set the current run.
        
        Args:
            run_id: ID of the run to set as current.
            
        Returns:
            Set Run object.
            
        Raises:
            TrackingError: If the run is not found.
        """
        run = self.storage.load_run(run_id)
        
        # If the run belongs to a different experiment, load and set that experiment
        if run.experiment_id and (not self.current_experiment or 
                                 run.experiment_id != self.current_experiment.experiment_id):
            try:
                self.current_experiment = self.storage.load_experiment(run.experiment_id)
            except TrackingError:
                # If the experiment doesn't exist, just set the run
                pass
        
        self.current_run = run
        
        self.logger.info(f"Set current run: {run.name} (ID: {run.run_id})")
        return run
    
    def list_runs(self, experiment_id: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        List all runs, optionally filtered by experiment ID.
        
        Args:
            experiment_id: Optional experiment ID to filter by.
            
        Returns:
            List of run metadata dictionaries.
        """
        if experiment_id is None and self.current_experiment:
            experiment_id = self.current_experiment.experiment_id
        
        return self.storage.list_runs(experiment_id)
    
    def delete_run(self, run_id: str) -> bool:
        """
        Delete a run.
        
        Args:
            run_id: ID of the run to delete.
            
        Returns:
            True if the run was deleted, False otherwise.
        """
        if self.current_run and self.current_run.run_id == run_id:
            self.current_run = None
        
        # If the run is part of the current experiment, remove it from there too
        if self.current_experiment:
            run = self.current_experiment.get_run(run_id)
            if run:
                self.current_experiment.runs.remove(run)
                self.storage.save_experiment(self.current_experiment)
        
        return self.storage.delete_run(run_id)
    
    def start_run(self) -> Run:
        """
        Start the current run.
        
        Returns:
            Started Run object.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.start()
        self.storage.save_run(self.current_run)
        
        self.logger.info(f"Started run: {self.current_run.name} (ID: {self.current_run.run_id})")
        return self.current_run
    
    def end_run(self, status: str = "completed", error_message: Optional[str] = None) -> Run:
        """
        End the current run.
        
        Args:
            status: Status of the run. Either "completed" or "failed".
            error_message: Optional error message if the run failed.
            
        Returns:
            Ended Run object.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        if status == "completed":
            self.current_run.complete()
        elif status == "failed":
            self.current_run.fail(error_message or "Run failed")
        else:
            raise TrackingError(f"Invalid run status: {status}. Must be 'completed' or 'failed'.")
        
        self.storage.save_run(self.current_run)
        
        self.logger.info(f"Ended run: {self.current_run.name} (ID: {self.current_run.run_id}) with status {status}")
        return self.current_run
    
    def log_parameter(self, key: str, value: Any) -> None:
        """
        Log a parameter for the current run.
        
        Args:
            key: Parameter name.
            value: Parameter value.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.log_parameter(key, value)
        self.storage.save_run(self.current_run)
    
    def log_parameters(self, params: Dict[str, Any]) -> None:
        """
        Log multiple parameters for the current run.
        
        Args:
            params: Dictionary of parameter names and values.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.log_parameters(params)
        self.storage.save_run(self.current_run)
    
    def log_metric(self, key: str, value: Union[int, float]) -> None:
        """
        Log a metric for the current run.
        
        Args:
            key: Metric name.
            value: Metric value.
            
        Raises:
            TrackingError: If no current run is set or the value is not a number.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.log_metric(key, value)
        self.storage.save_run(self.current_run)
    
    def log_metrics(self, metrics: Dict[str, Union[int, float]]) -> None:
        """
        Log multiple metrics for the current run.
        
        Args:
            metrics: Dictionary of metric names and values.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.log_metrics(metrics)
        self.storage.save_run(self.current_run)
    
    def log_metric_step(self, key: str, value: Union[int, float], step: int) -> None:
        """
        Log a metric with a step for the current run.
        
        Args:
            key: Metric name.
            value: Metric value.
            step: Step number.
            
        Raises:
            TrackingError: If no current run is set or the value is not a number.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.log_metric_step(key, value, step)
        self.storage.save_run(self.current_run)
    
    def log_artifact(self, key: str, artifact: Any, artifact_path: Optional[str] = None) -> str:
        """
        Log an artifact for the current run.
        
        Args:
            key: Artifact name.
            artifact: Artifact object.
            artifact_path: Optional path to save the artifact. If None, a path is generated.
            
        Returns:
            Path to the saved artifact.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        if artifact_path is None:
            # Generate a path in the artifacts directory
            artifacts_dir = os.path.join(self.storage.artifacts_dir, self.current_run.run_id)
            os.makedirs(artifacts_dir, exist_ok=True)
            
            artifact_path = os.path.join(artifacts_dir, f"{key}")
            
            # Add appropriate extension based on artifact type
            if hasattr(artifact, "savefig"):  # matplotlib figure
                artifact_path += ".png"
            elif isinstance(artifact, (dict, list)):
                artifact_path += ".json"
            else:
                artifact_path += ".pkl"
        
        location = self.current_run.log_artifact(key, artifact, artifact_path)
        self.storage.save_run(self.current_run)
        
        return location
    
    def log_model(self, model: Any, model_path: Optional[str] = None) -> str:
        """
        Log a model for the current run.
        
        Args:
            model: Model object.
            model_path: Optional path to save the model. If None, a path is generated.
            
        Returns:
            Path to the saved model.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        if model_path is None:
            # Generate a path in the artifacts directory
            artifacts_dir = os.path.join(self.storage.artifacts_dir, self.current_run.run_id)
            os.makedirs(artifacts_dir, exist_ok=True)
            
            model_path = os.path.join(artifacts_dir, "model.pkl")
        
        location = self.current_run.log_model(model, model_path)
        self.storage.save_run(self.current_run)
        
        return location
    
    def log_message(self, message: str) -> None:
        """
        Log a message for the current run.
        
        Args:
            message: Message to log.
            
        Raises:
            TrackingError: If no current run is set.
        """
        if not self.current_run:
            raise TrackingError("No current run set. Call create_run() or set_run() first.")
        
        self.current_run.log_message(message)
        self.storage.save_run(self.current_run)
    
    def add_metadata(self, key: str, value: Any, target: str = "run") -> None:
        """
        Add metadata to the current run or experiment.
        
        Args:
            key: Metadata key.
            value: Metadata value.
            target: Target for the metadata. Either "run" or "experiment".
            
        Raises:
            TrackingError: If no current run or experiment is set.
        """
        if target == "run":
            if not self.current_run:
                raise TrackingError("No current run set. Call create_run() or set_run() first.")
            
            self.current_run.add_metadata(key, value)
            self.storage.save_run(self.current_run)
        elif target == "experiment":
            if not self.current_experiment:
                raise TrackingError("No current experiment set. Call create_experiment() or set_experiment() first.")
            
            self.current_experiment.add_metadata(key, value)
            self.storage.save_experiment(self.current_experiment)
        else:
            raise TrackingError(f"Invalid target: {target}. Must be 'run' or 'experiment'.")
    
    def get_metadata(self, key: str, target: str = "run") -> Any:
        """
        Get metadata from the current run or experiment.
        
        Args:
            key: Metadata key.
            target: Target for the metadata. Either "run" or "experiment".
            
        Returns:
            Metadata value if found, None otherwise.
            
        Raises:
            TrackingError: If no current run or experiment is set.
        """
        if target == "run":
            if not self.current_run:
                raise TrackingError("No current run set. Call create_run() or set_run() first.")
            
            return self.current_run.get_metadata(key)
        elif target == "experiment":
            if not self.current_experiment:
                raise TrackingError("No current experiment set. Call create_experiment() or set_experiment() first.")
            
            return self.current_experiment.get_metadata(key)
        else:
            raise TrackingError(f"Invalid target: {target}. Must be 'run' or 'experiment'.")
    
    def get_best_run(self, metric: str, higher_is_better: bool = True):
        """
        Get the best run in the current experiment based on a metric.
        
        Args:
            metric: Name of the metric to use for comparison.
            higher_is_better: Whether higher values of the metric are better.
            
        Returns:
            Best Run object if found, None otherwise.
            
        Raises:
            TrackingError: If no current experiment is set or no runs have the specified metric.
        """
        if not self.current_experiment:
            raise TrackingError("No current experiment set. Call create_experiment() or set_experiment() first.")
        
        return self.current_experiment.get_best_run(metric, higher_is_better)
    
    def compare_runs(self, run_ids: List[str], metrics: List[str]) -> Dict[str, Dict[str, Any]]:
        """
        Compare runs based on specified metrics.
        
        Args:
            run_ids: List of run IDs to compare.
            metrics: List of metric names to compare.
            
        Returns:
            Dictionary mapping run IDs to dictionaries of metric values.
            
        Raises:
            TrackingError: If any of the runs are not found.
        """
        result = {}
        
        for run_id in run_ids:
            run = self.storage.load_run(run_id)
            
            run_metrics = {}
            for metric in metrics:
                if metric in run.metrics:
                    run_metrics[metric] = run.metrics[metric]
                else:
                    run_metrics[metric] = None
            
            result[run_id] = {
                "metrics": run_metrics,
                "parameters": run.parameters,
                "name": run.name,
                "status": run.status
            }
        
        return result
    
    def backup(self, backup_dir: Optional[str] = None) -> str:
        """
        Backup all tracking data to a directory.
        
        Args:
            backup_dir: Directory to backup to. If None, a directory in the current
                       working directory is created.
            
        Returns:
            Path to the backup directory.
        """
        if backup_dir is None:
            backup_dir = os.path.join(os.getcwd(), "sbyb_backups")
            os.makedirs(backup_dir, exist_ok=True)
        
        return self.storage.backup(backup_dir)
    
    def restore(self, backup_path: str) -> bool:
        """
        Restore tracking data from a backup.
        
        Args:
            backup_path: Path to the backup directory.
            
        Returns:
            True if the restore was successful, False otherwise.
        """
        success = self.storage.restore(backup_path)
        
        if success:
            # Reset current experiment and run
            self.current_experiment = None
            self.current_run = None
        
        return success
    
    def __enter__(self):
        """
        Enter context manager.
        
        Returns:
            Self.
        """
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit context manager.
        
        Args:
            exc_type: Exception type.
            exc_val: Exception value.
            exc_tb: Exception traceback.
        """
        if self.current_run and self.current_run.status == "running":
            if exc_type:
                # An exception occurred, mark the run as failed
                self.end_run(status="failed", error_message=str(exc_val))
            else:
                # No exception, mark the run as completed
                self.end_run(status="completed")
